from abc import ABC, abstractmethod

from django.conf import settings
from django.db.models import Q
from rest_framework import serializers

from arkindex.documents.models import Element, Transcription
from arkindex.project.tools import bounding_box, build_absolute_url


class AnnotationSerializer(ABC, serializers.BaseSerializer):
    """
    Serialize an element into a IIIF annotation
    """

    @abstractmethod
    def get_url(self, element):
        """Get an element's ID URL"""

    def get_resource(self, element):
        """Get the annotation resource for a given element."""
        return {
            "@id": self.get_url(element),
            "@type": "cnt:ContentAsText",
            "chars": element.name,
            "format": "text/plain"
        }

    def get_target(self, instance):
        """Get the target canvas (`on` property) for a given element."""
        if hasattr(instance, "element"):
            # A transcription zone may be defined by the element it is attached
            instance = instance.element
        assert instance.image_id is not None, "An image is required to get the target canvas of an element"
        assert instance.polygon is not None, "A polygon is required to get the target canvas of an element"

        url = instance.image.get_thumbnail_url(max_width=None, max_height=None)
        x, y, w, h = bounding_box(instance.polygon)
        return f"{url}#xywh={x},{y},{w},{h}"

    def to_representation(self, element):
        assert isinstance(element, (Element, Transcription))
        assert "request" in self.context, "A request is required to generate absolute URLs"
        return {
            "@id": self.get_url(element),
            "@type": "oa:Annotation",
            "motivation": "sc:painting",
            "on": self.get_target(element),
            "resource": self.get_resource(element),
        }


class TranscriptionAnnotationSerializer(AnnotationSerializer):
    """
    Serialize a transcription into a IIIF annotation
    """

    def get_url(self, ts):
        return build_absolute_url(
            ts,
            self.context["request"],
            "api:transcription-annotation",
            id_argument="page_pk",
            transcription_pk=ts.id,
        )

    def get_resource(self, ts):
        return {
            "@id": self.get_url(ts),
            "@type": "cnt:ContentAsText",
            "chars": ts.text,
            "format": "text/plain"
        }


class AnnotationListSerializer(serializers.Serializer):
    """
    Serialize a list of serialized annotations into a IIIF annotation list
    """

    annotation_serializer = AnnotationSerializer

    def to_representation(self, obj):
        assert "request" in self.context, "A request is required to generate absolute URLs"

        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": self.context["request"].build_absolute_uri(),
            "@type": "sc:AnnotationList",
            "resources": self.annotation_serializer(
                self.get_elements(obj),
                context=self.context,
                many=True
            ).data
        }

    def get_elements(self, element):
        """Get a list of elements to serialize as annotations."""
        return Transcription.objects.filter(
            Q(element=element)
            | Q(element__in=Element.objects.get_descending(element.id))
        )


class ElementAnnotationListSerializer(AnnotationListSerializer):
    """
    Serialize a page's transcriptions into a IIIF annotation list
    """

    annotation_serializer = TranscriptionAnnotationSerializer
