from unittest.mock import call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestMoveSelection(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.parent = cls.corpus.elements.get(name="Volume 2")
        cls.page = cls.corpus.elements.get(name="Volume 1, page 1r")

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(self.parent.id),
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(self.parent.id),
            })
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"selection": False})
    def test_disabled(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(self.parent.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(
                response.json(), {"non_field_errors": ["Selection is not available on this instance."]}
            )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_wrong_acl(self, filter_rights_mock):
        private_corpus = Corpus.objects.create(name="private", public=False)
        private_element = private_corpus.elements.create(
            type=private_corpus.types.create(slug="folder"),
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(private_corpus.id),
                "parent_id": str(private_element.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "corpus_id": [f'Invalid pk "{private_corpus.id}" - object does not exist.'],
                "parent_id": [f'Invalid pk "{private_element.id}" - object does not exist.']
            }
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_wrong_parent(self):
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.page)
        with self.assertNumQueries(4):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": "12341234-1234-1234-1234-123412341234"
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"parent_id": ['Invalid pk "12341234-1234-1234-1234-123412341234" - object does not exist.']}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_same_parent(self):
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.page)
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(self.page.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"parent_id": ["An element cannot be added as a parent of itself."]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_move_element_different_corpus(self):
        corpus2 = Corpus.objects.create(name="new")
        corpus2.memberships.create(user=self.user, level=Role.Contributor.value)
        parent = corpus2.elements.create(type=corpus2.types.create(slug="folder"))
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.page)
        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(parent.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"parent_id": ["An element cannot be added to a parent from another corpus."]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_parent_is_direct_parent(self):
        parent = self.corpus.elements.get(name="Volume 1")
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.page)
        with self.assertNumQueries(7):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(parent.id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"parent_id": [f"'{parent.id}' is already a direct parent of one or more selected elements."]}
        )

    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_parent_is_child(self):
        target = self.corpus.elements.get(name="Volume 1")
        parent_id = self.page.id
        self.client.force_login(self.user)
        self.user.selected_elements.add(target)
        with self.assertNumQueries(8):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(parent_id),
            })
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"parent_id": [f"'{parent_id}' is a child of one or more selected elements."]}
        )

    @patch("arkindex.project.triggers.documents_tasks.add_parent_selection.delay")
    @override_settings(ARKINDEX_FEATURES={"selection": True})
    def test_delay(self, delay_mock):
        self.client.force_login(self.user)
        self.user.selected_elements.add(self.page)
        another_page = self.corpus.elements.get(name="Volume 1, page 1v")
        self.user.selected_elements.add(another_page)
        with self.assertNumQueries(8):
            response = self.client.post(reverse("api:create-parent-selection"), {
                "corpus_id": str(self.corpus.id),
                "parent_id": str(self.parent.id),
            })
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            parent=self.parent,
            user_id=self.user.id,
            description=f"Adding {self.parent.name} as parent of selected elements from corpus {self.corpus.id}"
        ))
