from collections import defaultdict
from textwrap import dedent

from django.conf import settings
from django.utils.module_loading import import_string
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied, ValidationError

from arkindex.documents.models import Corpus, Element, ElementType, MLClass
from arkindex.ponos.models import Farm, State
from arkindex.ponos.serializers import TaskLightSerializer
from arkindex.process.models import (
    ActivityState,
    DataFile,
    FeatureUsage,
    Process,
    ProcessMode,
    WorkerRun,
    WorkerVersionState,
)
from arkindex.process.utils import get_default_farm
from arkindex.project.mixins import ProcessACLMixin
from arkindex.project.serializer_fields import EnumField, LinearRingField
from arkindex.project.validators import MaxValueValidator
from arkindex.training.models import ModelVersionState
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level

ProcessFarmField = import_string(getattr(settings, "PROCESS_FARM_FIELD", None) or "arkindex.project.serializer_fields.NullField")


class ProcessLightSerializer(serializers.ModelSerializer):
    """
    Serialize a process
    """

    state = EnumField(State, required=False)
    mode = EnumField(ProcessMode, read_only=True)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())
    activity_state = EnumField(ActivityState, read_only=True)

    class Meta:
        model = Process
        fields = (
            "name",
            "id",
            "state",
            "mode",
            "corpus",
            "creator",
            "activity_state",
            "use_cache",
            "chunks",
        )
        read_only_fields = ("id", "state", "mode", "corpus", "creator", "activity_state", "use_cache", "chunks",)


class CorpusWorkerRunProcessSerializer(ProcessLightSerializer):
    element_id = serializers.PrimaryKeyRelatedField(read_only=True, allow_null=True)
    element_count = serializers.IntegerField(
        read_only=True,
        allow_null=True,
        help_text=dedent("""
            This count is only set for processes created from a selection of elements,
            and does not account for either process filters nor the recursive option.
        """)
    )

    class Meta(ProcessLightSerializer.Meta):
        fields = ProcessLightSerializer.Meta.fields + (
            "element_id",
            "element_count"
        )


class ProcessSerializer(ProcessLightSerializer):
    """
    Serialize a process with its settings
    """
    from arkindex.documents.serializers.elements import ElementTinySerializer
    element = ElementTinySerializer(read_only=True)
    element_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        default=None,
        allow_null=True,
        write_only=True,
        source="element",
        style={"base_template": "input.html"},
    )
    folder_type = serializers.SlugRelatedField(slug_field="slug", default=None, read_only=True)
    element_type = serializers.SlugRelatedField(
        queryset=ElementType.objects.none(),
        slug_field="slug",
        allow_null=True,
        style={"base_template": "input.html"},
    )
    element_name_contains = serializers.CharField(
        source="name_contains",
        allow_null=True,
        allow_blank=True,
        max_length=250
    )
    ml_class_id = serializers.PrimaryKeyRelatedField(
        queryset=MLClass.objects.none(),
        default=None,
        allow_null=True,
        source="ml_class",
        style={"base_template": "input.html"},
    )

    template_id = serializers.PrimaryKeyRelatedField(
        queryset=Process.objects.none(),
        allow_null=True,
        source="template",
        style={"base_template": "input.html"}
    )

    class Meta(ProcessLightSerializer.Meta):
        fields = ProcessLightSerializer.Meta.fields + (
            "files",
            "element",
            "element_id",
            "folder_type",
            "element_type",
            "element_name_contains",
            "ml_class_id",
            "load_children",
            "template_id",
        )
        read_only_fields = ProcessLightSerializer.Meta.read_only_fields + (
            "files",
            "element",
            "folder_type",
            "template_id",
        )


class ProcessDetailsSerializer(ProcessSerializer):
    farm = ProcessFarmField(read_only=True)
    tasks = TaskLightSerializer(many=True, read_only=True)

    class Meta(ProcessSerializer.Meta):
        fields = ProcessSerializer.Meta.fields + (
            "farm",
            "tasks",
            "created",
            "updated",
            "started",
            "finished",
        )
        read_only_fields = (
            # Make the state writable again to allow stopping processes
            tuple(set(ProcessSerializer.Meta.read_only_fields) - {"state"}) + (
                "farm",
                "tasks",
                "created",
                "updated",
                "started",
                "finished",
            )
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.instance or not self.instance.corpus:
            return

        self.fields["element_type"].queryset = self.instance.corpus.types.all()
        self.fields["element_id"].queryset = (
            self.instance.corpus.elements.all()
            # Include the extra attributes that will be serialized in the response
            .select_related("type", "image__server")
        )
        self.fields["ml_class_id"].queryset = self.instance.corpus.ml_classes.all()

    def validate(self, data):
        data = super().validate(data)
        edited_fields = set(data.keys())

        # Fields that can always be edited on any process of any state
        editable_fields = {"name", "state"}

        # Allow editing the element ID and name on Files processes at any time
        # TODO: Only allow editing the element ID on a running file import to Ponos tasks,
        # since this edition is only permitted to show the "View element" button once the import completes.
        if self.instance.mode == ProcessMode.Files:
            editable_fields.add("element")

        # If some fields are being edited that are not the fields that are always editable
        if edited_fields - editable_fields:
            # Fail if the process is running or about to run
            if self.instance.state in (State.Pending, State.Running):
                raise serializers.ValidationError({"__all__": ["Cannot edit a process while it is running"]})

            # Otherwise, only allow editing Workers processes.
            if self.instance.mode != ProcessMode.Workers:
                raise serializers.ValidationError({"__all__": ["Only processes of mode Workers can be updated"]})

        return data

    def validate_state(self, state):
        if state != State.Stopping:
            raise serializers.ValidationError("Can only change the state to 'stopping'")

        if self.instance.state not in (State.Unscheduled, State.Pending, State.Running):
            raise serializers.ValidationError(f"Cannot stop a {self.instance.state.value} process")

        return state

    def update(self, instance, validated_data):
        if validated_data.get("state") == State.Stopping:
            self.instance.stop()
            # This is a read-only property of a Process, so we have to remove it to prevent it from being set by DRF
            del validated_data["state"]

        return super().update(instance, validated_data)


class ProcessListSerializer(ProcessLightSerializer):
    creator = serializers.CharField(
        source="creator.display_name",
        read_only=True,
        help_text="Display name of the user who created the process.",
    )

    class Meta(ProcessLightSerializer.Meta):
        fields = ProcessLightSerializer.Meta.fields + ("created", "updated", "started", "finished")
        read_only_fields = ProcessLightSerializer.Meta.read_only_fields + ("created", "updated", "started", "finished")
        # Show in the API docs that Process.started and Process.finished are nullable; this is not detected automatically
        extra_kwargs = {
            "started": {"allow_null": True},
            "finished": {"allow_null": True},
        }


class FilesProcessSerializer(serializers.ModelSerializer):

    files = serializers.PrimaryKeyRelatedField(
        queryset=DataFile.objects.select_related("corpus"),
        many=True,
    )
    folder_id = serializers.UUIDField(
        required=False,
        allow_null=True,
        source="element",
    )
    folder_type = serializers.SlugField()
    element_type = serializers.SlugField()
    farm_id = serializers.PrimaryKeyRelatedField(
        queryset=Farm.objects.all(),
        allow_null=True,
        default=None,
        source="farm",
    )

    # Automatically set the creator on the process
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())

    default_error_messages = {
        "files_required": "At least one file is required to start an import process",
        "unique_corpus": "Imports can only run on files from a single corpus",
        "corpus_read_only": "Cannot write in corpus",
        "folder_not_found": "Folder does not exist",
        "unsupported_content_type": "File imports can only import images, PDF documents, ZIP archives or IIIF manifests",
        "type_not_found": "Element type {slug!r} does not exist",
        "type_not_folder": "Element type {slug!r} is not a folder",
        "type_folder": "Element type {slug!r} is a folder",
        "wrong_folder_corpus": "Element and files are in different corpora",
    }

    class Meta:
        model = Process
        fields = (
            "files",
            "folder_id",
            "folder_type",
            "element_type",
            "farm_id",
            "creator",
        )

    def validate_files(self, files):
        if not files:
            self.fail("files_required")
        corpora = set(f.corpus for f in files)
        if len(corpora) > 1:
            self.fail("unique_corpus")
        corpus = corpora.pop()
        if not Corpus.objects.writable(self.context["request"].user).filter(id=corpus.id).exists():
            self.fail("corpus_read_only")
        return files

    def validate_folder_id(self, folder_id):
        if not folder_id:
            return None
        try:
            return Element.objects.get(
                id=folder_id,
                type__folder=True,
            )
        except Element.DoesNotExist:
            self.fail("folder_not_found")

    def validate_farm_id(self, farm):
        if farm is None:
            farm = get_default_farm()

        if farm and not farm.is_available(self.context["request"].user):
            raise ValidationError(["You do not have access to this farm."])

        return farm

    def validate(self, data):
        if not all(
            f.content_type == "application/pdf"
            or f.content_type in settings.ARCHIVE_MIME_TYPES
            or f.content_type.startswith("image/")
            or f.content_type.split(";")[0] in ["application/json", "application/ld+json"]
            for f in data["files"]
        ):
            self.fail("unsupported_content_type")

        data["corpus"] = corpus = data["files"][0].corpus

        element_type_slug = data.get("element_type")
        element_type = corpus.types.filter(slug=element_type_slug).first()
        if not element_type:
            self.fail("type_not_found", slug=element_type_slug)
        if element_type.folder:
            self.fail("type_folder", slug=element_type_slug)

        data["element_type"] = element_type
        folder_type_slug, folder = data.get("folder_type"), data.get("element")

        folder_type = corpus.types.filter(slug=folder_type_slug).first()
        if not folder_type:
            self.fail("type_not_found", slug=folder_type_slug)
        if not folder_type.folder:
            self.fail("type_not_folder", slug=folder_type_slug)
        data["folder_type"] = folder_type
        # When a folder is set, it must be in the same corpus as all files
        if folder and folder.corpus_id != corpus.id:
            self.fail("wrong_folder_corpus")

        return data

    def create(self, validated_data):
        files = validated_data.pop("files")
        process = super().create({"mode": ProcessMode.Files.value, **validated_data})
        process.files.set(files)
        process.run()
        return process


class StartProcessSerializer(serializers.Serializer):
    chunks = serializers.IntegerField(
        min_value=1,
        validators=[MaxValueValidator(lambda: settings.MAX_CHUNKS)],
        default=1,
    )
    farm = serializers.PrimaryKeyRelatedField(queryset=Farm.objects.all(), default=None, allow_null=True)
    use_cache = serializers.BooleanField(default=False)
    worker_activity = serializers.BooleanField(default=False)

    def validate_farm(self, farm):
        """
        Using the `default=` on the farm field would allow the default farm
        to be used without having access rights to it, so we do the ACL checks
        here during validation. We thus do not filter the field's queryset,
        as we would be duplicating the ACL checks otherwise.
        """
        if farm is None:
            farm = get_default_farm()

        if farm and not farm.is_available(self.context["request"].user):
            raise ValidationError(["You do not have access to this farm."])

        return farm

    def validate(self, validated_data):
        assert self.instance is not None, "A Process instance is required for this serializer"
        errors = defaultdict(list)

        if self.instance.mode == ProcessMode.Dataset:
            # Only call .count() and .all() as they will access the prefetched dataset sets and not cause any extra query
            if not self.instance.sets.count():
                errors["non_field_errors"].append("A dataset process cannot be started if it does not have any associated dataset sets.")
            elif not any(ds.dataset.corpus_id == self.instance.corpus.id for ds in self.instance.sets.all()):
                errors["non_field_errors"].append("At least one of the process sets must be from the same corpus as the process.")

            if validated_data.get("worker_activity"):
                errors["worker_activity"].append("Worker activities are not supported on Dataset processes.")
            if validated_data.get("use_cache"):
                errors["use_cache"].append("Caching is not supported on Dataset processes.")

        # Use process.worker_runs.all() to access the (prefetched) worker_runs to avoid new SQL queries
        # The related worker versions have also been prefetched
        if len(list(self.instance.worker_runs.all())) > 0:
            # Check if a worker run has no model version but version.model_usage = Required
            missing_model_versions = []
            unavailable_versions = []
            has_unavailable_model_versions = False
            missing_required_configurations = []

            for worker_run in self.instance.worker_runs.all():
                if worker_run.version.model_usage == FeatureUsage.Required and worker_run.model_version_id is None:
                    missing_model_versions.append(worker_run.version.worker.name)
                if (
                    worker_run.version.state != WorkerVersionState.Available
                    or not worker_run.version.docker_image_iid
                ):
                    unavailable_versions.append(worker_run.version.id)
                if worker_run.model_version_id and worker_run.model_version.state != ModelVersionState.Available:
                    has_unavailable_model_versions = True

                # If the worker version has a user configuration, check that we aren't missing any required fields
                if isinstance(worker_run.version.configuration.get("user_configuration"), dict):
                    required_fields = set(
                        name
                        for name, options in worker_run.version.configuration["user_configuration"].items()
                        # Only pick the fields that are required and without default values
                        if options.get("required") and "default" not in options
                    )

                    # List all the fields defined on the WorkerRun's configuration if there is one
                    worker_run_fields = set()
                    if worker_run.configuration is not None:
                        worker_run_fields = set(worker_run.configuration.configuration.keys())

                    # Check that all the worker version's required fields are set in that configuration
                    if not required_fields.issubset(worker_run_fields):
                        missing_required_configurations.append(worker_run.version.worker.name)

            if len(missing_model_versions) > 0:
                errors["model_version"].append(f"The following workers require a model version and none was set: {missing_model_versions}")
            if has_unavailable_model_versions:
                errors["model_version"].append("This process contains one or more unavailable model versions and cannot be started.")
            if len(unavailable_versions) > 0:
                errors["version"].append("This process contains one or more unavailable worker versions and cannot be started.")
            if missing_required_configurations:
                errors["worker_configuration"].append(
                    f'The following workers have required configuration fields that have not been set: {", ".join(sorted(missing_required_configurations))}',
                )

        else:
            errors["__all__"].append("The process must have worker runs to be started.")

        if errors:
            raise ValidationError(errors)

        # The Process model uses activity_state, but we have a boolean in the serializer, so we have to convert
        validated_data["activity_state"] = ActivityState.Pending if validated_data.pop("worker_activity") else ActivityState.Disabled

        return validated_data

    def update(self, process, validated_data):
        for attr, value in validated_data.items():
            setattr(process, attr, value)
        process.run()
        return process


class CreateProcessTemplateSerializer(serializers.ModelSerializer):
    name = serializers.CharField(required=False, max_length=250, default="")

    class Meta:
        model = Process
        fields = (
            "id",
            "name",
        )

    def validate(self, data):
        data = super().validate(data)
        template = self.context["template"]
        data["mode"] = ProcessMode.Template
        data["creator"] = self.context["request"].user
        data["corpus"] = template.corpus

        # Can only be "disabled" or "pending.
        data["activity_state"] = ActivityState.Disabled
        return data

    def create(self, validated_data):
        # The endpoint should copy all the worker runs content from the source template
        # (from url id) onto the newly created one.
        process = super().create(validated_data)
        self.context["template"].copy_runs(process)
        return process


class ApplyProcessTemplateSerializer(ProcessACLMixin, serializers.Serializer):
    process_id = serializers.PrimaryKeyRelatedField(
        queryset=Process.objects.select_related("corpus"),
        source="process",
        required=True,
        help_text="ID of the process to apply the template to",
    )

    def create(self, validated_data):
        # The endpoint should copy all the worker runs content from the source template
        # (from url id) onto the newly created one.
        template_process = self.context["template"]
        target_process = validated_data["process"]
        # If the target process, already has worker runs, these will be deleted before applying the template
        WorkerRun.objects.filter(process_id=target_process.id).delete()
        # Apply the template by copying all the worker runs on to the new process
        template_process.copy_runs(target_process)
        target_process.template_id = template_process.id
        target_process.save(update_fields=["template_id"])
        return target_process

    def validate_process_id(self, process):
        # Check Contributor access rights on process, and show only the ACL errors before any other error
        access_level = self.process_access_level(process)
        if not access_level:
            raise ValidationError(detail="Process with this ID does not exist.")
        if access_level < Role.Contributor.value:
            raise PermissionDenied(detail="You do not have a contributor access to this process.")

        errors = []
        if process.template_id is not None:
            errors.append("A template has already been applied to this process.")
        if process.mode not in (ProcessMode.Workers, ProcessMode.Dataset):
            errors.append("Templates can only be applied on Workers or Dataset processes.")
        if errors:
            raise ValidationError(errors)

        return process

    def validate(self, data):
        template_process = self.context["template"]

        # The template's worker runs have already been prefetched, so we can use them to validate directly
        if any(
            run.version.state != WorkerVersionState.Available
            or not run.version.docker_image_iid
            for run in template_process.worker_runs.all()
        ):
            raise ValidationError(detail="This template contains one or more unavailable worker versions and cannot be applied.")

        if any(
            run.version.worker.archived
            for run in template_process.worker_runs.all()
        ):
            raise ValidationError(detail="This template contains one or more worker versions from archived workers and cannot be applied.")

        if any(
            run.model_version.state != ModelVersionState.Available
            for run in template_process.worker_runs.all()
            if run.model_version_id is not None
        ):
            raise ValidationError(detail="This template contains one or more unavailable model versions and cannot be applied.")

        if any(
            run.model_version.model.archived
            for run in template_process.worker_runs.all()
            if run.model_version_id is not None
        ):
            raise ValidationError(detail="This template contains one or more model versions from archived models and cannot be applied.")

        return data


class CorpusProcessSerializer(serializers.Serializer):
    # TODO: Rewrite this as a ModelSerializer and merge the validation code from the API into it

    corpus = serializers.UUIDField()
    name = serializers.CharField(required=False, max_length=250)
    mode = serializers.ChoiceField([ProcessMode.Workers.value, ProcessMode.Dataset.value], default=ProcessMode.Workers.value)

    # Elements filtering options
    element = serializers.UUIDField(required=False)
    element_name_contains = serializers.CharField(required=False, max_length=250)
    element_type = serializers.SlugField(required=False, max_length=50)

    ml_class_id = serializers.PrimaryKeyRelatedField(
        queryset=MLClass.objects.all(),
        required=False,
        style={"base_template": "input.html"},
        help_text="Only include elements that have a classification of a specified MLClass",
        # Expose the ID as ml_class_id, but store the MLClass from this field in the correct attribute
        source="ml_class",
    )

    selection = serializers.BooleanField(default=False)
    load_children = serializers.BooleanField(default=False)
    use_cache = serializers.BooleanField(default=False)

    def validate_corpus(self, corpus_id):
        try:
            corpus = Corpus.objects.get(id=corpus_id)
        except Corpus.DoesNotExist:
            raise ValidationError(["Corpus with this ID does not exist."])

        # Assert user has an admin access to the corpus
        access_level = get_max_level(self.context["request"].user, corpus)
        if not access_level:
            raise ValidationError(["Corpus with this ID does not exist."])
        if access_level < Role.Admin.value:
            raise ValidationError(["You do not have an admin access to this corpus."])
        return corpus

    def validate_element(self, element_id):
        if not element_id:
            return None
        try:
            return Element.objects.select_related("type").get(id=element_id)
        except Element.DoesNotExist:
            raise ValidationError(["Element with this ID does not exist."])

    def validate(self, data):
        """
        Handle filters validation
        """
        corpus = data.get("corpus")

        element = data.get("element")
        elt_type = data.get("element_type")
        ml_class = data.get("ml_class")
        selection = data.get("selection")
        mode = ProcessMode(data.get("mode"))

        # TODO: Assemble the errors into a defaultdict to raise multiple errors at once
        if selection and not settings.ARKINDEX_FEATURES["selection"]:
            raise ValidationError({"selection": ["Selection is not available on this instance."]})

        if selection and element:
            raise ValidationError({
                "element": ["This field cannot be set with the selection field."],
                "selection": ["This field cannot be set with the element field."]
            })

        if selection and mode == ProcessMode.Dataset:
            raise ValidationError({
                "mode": ["This process mode cannot be set with the selection field."],
                "selection": ["This field cannot be set with the specified process mode."]
            })

        # If element is defined ensure it is part of the right corpus
        if element and element.corpus_id != corpus.id:
            raise ValidationError({
                "element": [f"Element is not part of corpus {corpus.name}"]
            })

        # Check type filter is valid
        if elt_type:
            try:
                data["element_type"] = corpus.types.get(slug=elt_type)
            except ElementType.DoesNotExist:
                raise ValidationError({
                    "element_type": [f'This type does not exist in corpus "{corpus.name}"']
                })

        if ml_class and ml_class.corpus_id != corpus.id:
            raise ValidationError({
                "ml_class_id": ["This MLClass is not in the selected corpus."],
            })

        return data


class ProcessElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises an Element, using optimized query for ListProcessElement
    """
    class Meta:
        model = Element
        fields = (
            "id",
            "type_id",
            "name",
            "confidence",
        )
        read_only_fields = fields


class ProcessElementSerializer(ProcessElementLightSerializer):
    image_id = serializers.UUIDField(allow_null=True)
    image_width = serializers.IntegerField(min_value=0, source="image__width", allow_null=True)
    image_height = serializers.IntegerField(min_value=0, source="image__height", allow_null=True)
    image_url = serializers.URLField(allow_null=True)
    polygon = LinearRingField(allow_null=True)
    rotation_angle = serializers.IntegerField(min_value=0, max_value=359, default=0)
    mirrored = serializers.BooleanField(default=False)

    class Meta(ProcessElementLightSerializer.Meta):
        fields = ProcessElementLightSerializer.Meta.fields + (
            "image_id",
            "image_width",
            "image_height",
            "image_url",
            "polygon",
            "rotation_angle",
            "mirrored"
        )
        read_only_fields = fields


class ProcessFailuresSerializer(ProcessSerializer):
    # Override state field from ProcessLightSerializer to mark it as read-only
    state = EnumField(State, required=False, read_only=True)

    def validate(self, data):
        base_process = self.context["process"]
        errors = defaultdict(list)
        if base_process.mode != ProcessMode.Workers:
            errors["__all__"].append("The process must be in mode Workers to perform this action")
        if not base_process.finished:
            errors["__all__"].append("The process must be finished to perform this action")
        if not base_process.has_failures:
            errors["__all__"].append("The process has no element with a WorkerActivity in an Error state")
        if errors:
            raise ValidationError(errors)
        return data

    class Meta(ProcessLightSerializer.Meta):
        read_only_fields = ProcessSerializer.Meta.read_only_fields + (
            "name",
            "state",
            "element_type",
            "element_name_contains",
            "ml_class_id",
            "load_children",
        )
